#!/usr/bin/python3

# Shared Skies focuser control
# Copyright (c) 2012-2018 John Kielkopf
# kielkopf@louisville.edu

# This file is part of the PyFocus package for telescope operations

# Date: May 17, 2018
# Version: 2.1


#
# Changes
#

# 2018-04-05
# Version 2.0 updated from 1.0 to Python 3 and Phidget22

# 2018-05-17
# Version 2.1 added tkinter GUI


#Basic imports

import sys
from time import sleep       # for delays 


#Phidget imports

from Phidget22.Devices.Encoder import *
from Phidget22.Devices.DCMotor import *
from Phidget22.PhidgetException import *
from Phidget22.Phidget import *


# Tkinter imports

import tkinter as tk
from tkinter import ttk

#
# Change these for specific telescope/focuser/motor combination
#

# Diagnostics flag

diagnostics = False

# Location of status file

#focus_status_file = "/usr/local/observatory/status/telfocus"

focus_status_file = './telfocus'

#
# Begin user installation  parameters
#

#
# Celestron motor on the CDK20 telescopes
#
# Motor RPM at 100% duty cycle: 4000.
# Motor RPM speed factor: 0.5 estimated
# Motor shaft encoder counts/turn: 360.
# Motor gear reduction: 139. + 184./1221. = 139.150696151
# Focuser screw pitch: 32 tpi
# Focuser translation scale: 793.75 microns / shaft revolution
# Motor counts per micron focuser motion: 63.111
# Focal ratio: f/6.8
# Pixel size: 9 microns
# Secondary motion magnification: 1.0 
#
# Set CDK20 motor and focuser parameters
#
# counts_per_revolution = 360.
# revolutions_per_second = 0.5*4000./60.
# counts_per_second = counts_per_revolution * revolutions_per_second
# counts_per_micron = 63.11
# dc_max = 1000.*counts_per_micron
#
# Set CDK20 telescope parameters
#
# f_ratio = 6.8
# f_mag = 1.0
# pixel_size = 9.0
# arcsecond_per_pixel = 0.53
#
#
# Pittman 9237S011-R1 motor on the RC telescopes
#
# Set RC motor and focuser parameters
#
# Screw pitch: 20 tpi
# Motor encoder: 500 counts per turn
# Motor gearing: 1:1
# No-load speed: 4370 rpm 
#
# counts_per_revolution = 500.
# revolutions_per_second = 20.
# counts_per_second = counts_per_revolution * revolutions_per_second
# counts_per_micron = 0.3937
# dc_max = 1000.*counts_per_micron
#
# Set RC telescope parameters
#
# f_ratio = 8.0
# f_mag = 4.0
# pixel_size = 9.0
# arcsecond_per_pixel = 0.39


# Set motor and focuser parameters

counts_per_revolution = 360.
revolutions_per_second = 0.5*4000./60.
counts_per_second = counts_per_revolution * revolutions_per_second
counts_per_micron = 63.11
dc_max = 1000.*counts_per_micron

# Set telescope parameters

f_ratio = 6.8
f_mag = 1.0
pixel_size = 9.0
arcsecond_per_pixel = 0.53

# Set image scale for pixel and arcseond defocus units

counts_per_pixel =  pixel_size*f_ratio*counts_per_micron/f_mag
counts_per_arcsec = counts_per_pixel/arcsecond_per_pixel

#
# End of user installation parameters
#



# These  hold the  current settings

global motor_count
global motor_velocity
global focus_position
global focus_image
global focus_sky

# These hold the incremental changes

global change_counts
global change_position
global change_image
global change_sky

# Initialize changes that will be updated by the GUI

change_counts = 0
change_position = 0.
change_image = 0.
change_sky = 0.
 
# Intialize values that will be updated by encoder interrupt

motor_count = 0
motor_velocity = 0


# Look up the saved last encoder reading of the focuserif any
# Assign current encoder count to match or else set to zero

try: 
  focus_status_fp = open(focus_status_file,'r')
  focus_str = focus_status_fp.read(40)
  focus_status_fp.close()
  motor_count = float(focus_str)
except:
  print (focus_status_file+" is missing or empty")
  print ("Setting focus position to zero")
  print (" ")
  motor_count = 0.

# Update focus information from motor count

focus_position = motor_count/counts_per_micron
focus_image =abs( motor_count)/counts_per_pixel
focus_sky = abs(motor_count)/counts_per_arcsec  


# Setup the motor

try:
  ch_motor = DCMotor()
except RuntimeError as e:
  print("Runtime Exception %s" % e.details)
  print("Press Enter to Exit...\n")
  readin = sys.stdin.read(1)
  exit(1)

def DCMotorAttached(e):
  try:
    attached = e
    #print("\nAttach Event Detected (Information Below)")
    #print("===========================================")
    #print("Library Version: %s" % attached.getLibraryVersion())
    #print("Serial Number: %d" % attached.getDeviceSerialNumber())
    #print("Channel: %d" % attached.getChannel())
    #print("Channel Class: %s" % attached.getChannelClass())
    #print("Channel Name: %s" % attached.getChannelName())
    #print("Device ID: %d" % attached.getDeviceID())
    #print("Device Version: %d" % attached.getDeviceVersion())
    #print("Device Name: %s" % attached.getDeviceName())
    #print("Device Class: %d" % attached.getDeviceClass())
    #print("\n")

  except PhidgetException as e:
    print("Phidget Exception %i: %s" % (e.code, e.details))
    #print("Press Enter to Exit...\n")
    readin = sys.stdin.read(1)
    exit(1)   
  
def DCMotorDetached(e):
  detached = e
  try:
    detached_port = detached.getHubPort()
    detached_channel = detached.getChannel()
    #print("\nDetach event on Port %d Channel %d" % (detached_port, detached_channel))
  except PhidgetException as e:
    print("Phidget Exception %i: %s" % (e.code, e.details))
    print("Press Enter to Exit...\n")
    readin = sys.stdin.read(1)
    exit(1)   

def ErrorEvent(e, eCode, description):
  print("Error %i : %s" % (eCode, description))

def VelocityUpdateHandler(e, velocity):
  global motor_velocity
  motor_velocity = velocity
  #print("Velocity: %f" % velocity)

try:
  ch_motor.setOnAttachHandler(DCMotorAttached)
  ch_motor.setOnDetachHandler(DCMotorDetached)
  ch_motor.setOnErrorHandler(ErrorEvent)

  ch_motor.setOnVelocityUpdateHandler(VelocityUpdateHandler)

  #print("Waiting for the Phidget DCMotor Object to be attached...")
  ch_motor.openWaitForAttachment(5000)
except PhidgetException as e:
  print("Phidget Exception %i: %s" % (e.code, e.details))
  print("Press Enter to Exit...\n")
  readin = sys.stdin.read(1)
  exit(1)


# Set up the encoder

try:
  ch_encoder = Encoder()
except RuntimeError as e:
  print("Runtime Exception %s" % e.details)
  print("Press Enter to Exit...\n")
  readin = sys.stdin.read(1)
  exit(1)

def EncoderAttached(self):
  try:
    attached = self
    #print("\nAttach Event Detected (Information Below)")
    #print("===========================================")
    #print("Library Version: %s" % attached.getLibraryVersion())
    #print("Serial Number: %d" % attached.getDeviceSerialNumber())
    #print("Channel: %d" % attached.getChannel())
    #print("Channel Class: %s" % attached.getChannelClass())
    #print("Channel Name: %s" % attached.getChannelName())
    #print("Device ID: %d" % attached.getDeviceID())
    #print("Device Version: %d" % attached.getDeviceVersion())
    #print("Device Name: %s" % attached.getDeviceName())
    #print("Device Class: %d" % attached.getDeviceClass())
    #print("\n")

  except PhidgetException as e:
    print("Phidget Exception %i: %s" % (e.code, e.details))
    print("Press Enter to Exit...\n")
    readin = sys.stdin.read(1)
    exit(1)   
  
def EncoderDetached(self):
  detached = self
  try:
    detached_port = detached.getHubPort()
    detached_channel = detached.getChannel()
    #print("\nDetach event on Port %d Channel %d" % (detached_port, detached_channel))
  except PhidgetException as e:
    print("Phidget Exception %i: %s" % (e.code, e.details))
    print("Press Enter to Exit...\n")
    readin = sys.stdin.read(1)
    exit(1)   

def ErrorEvent(self, eCode, description):
  print("Error %i : %s" % (eCode, description))

def PositionChangeHandler(self, positionChange, timeChange, indexTriggered):
  global motor_count
  
  # Update the encoder count and focus values
  
  motor_count += positionChange    
  

try:
  ch_encoder.setOnAttachHandler(EncoderAttached)
  ch_encoder.setOnDetachHandler(EncoderDetached)
  ch_encoder.setOnErrorHandler(ErrorEvent)
  ch_encoder.setOnPositionChangeHandler(PositionChangeHandler)
  ch_encoder.openWaitForAttachment(5000)
except PhidgetException as e:
  print("Phidget Exception %i: %s" % (e.code, e.details))
  print("Press Enter to Exit...\n")
  readin = sys.stdin.read(1)
  exit(1)

if (not ch_encoder.getEnabled()):
  ch_encoder.setEnabled(1)

#
# Done with set up of parameters, file, motor and encoder
#




#
# Build the GUI
#


# Callbacks for buttons
  
def focus_in():

  global change_counts
  global focus_position
  global focus_image
  global focus_sky

  request = -1.*change_counts
  result = change_focus(request)
    
  if diagnostics:
    print ("Requested change (counts): ", request)
    print ("Actual change (counts): ", result)

  # Update focus information from motor rotation
  
  focus_position = motor_count/counts_per_micron
  focus_image =abs( motor_count)/counts_per_pixel
  focus_sky = abs(motor_count)/counts_per_arcsec  

  # Update GUI
  
  tk_focus_position.set(format(focus_position, '7.0f'))
  tk_focus_image.set(format(focus_image, '7.1f'))
  tk_focus_sky.set(format(focus_sky, '7.1f'))

  # Write the status file with the encoder count

  focus_status_fp = open(focus_status_file,'w')
  focus_status_fp.write("%7.0f \n" % motor_count)
  focus_status_fp.close()

        
  return()

def focus_out():
  
  global change_counts
  global focus_position
  global focus_image
  global focus_sky

  request = change_counts
  result = change_focus(request)
  
  if diagnostics:
    print ("Requested change (counts): ", request)
    print ("Actual change (counts): ", result)

  # Update focus information from motor rotation
  
  focus_position = motor_count/counts_per_micron
  focus_image = abs(motor_count)/counts_per_pixel
  focus_sky = abs(motor_count)/counts_per_arcsec  


  # Update GUI
  
  tk_focus_position.set(format(focus_position, '7.0f'))
  tk_focus_image.set(format(focus_image, '7.1f'))
  tk_focus_sky.set(format(focus_sky, '7.1f'))

  # Write the status file with the encoder count

  focus_status_fp = open(focus_status_file,'w')
  focus_status_fp.write("%f \n" % motor_count)
  focus_status_fp.close()

  return()

def focus_reset():

  global motor_count
  global focus_position
  global focus_image
  global focus_sky
  
  if diagnostics:
    print ("Requested reset of count : ", motor_count)

  motor_count = 0
  
  # Update focus information from motor rotation
  
  focus_position = motor_count/counts_per_micron
  focus_image = abs(motor_count)/counts_per_pixel
  focus_sky = abs(motor_count)/counts_per_arcsec  


  # Update GUI
  
  tk_focus_position.set(format(focus_position, '7.0f'))
  tk_focus_image.set(format(focus_image, '7.1f'))
  tk_focus_sky.set(format(focus_sky, '7.1f'))

  # Write the status file with the encoder count

  focus_status_fp = open(focus_status_file,'w')
  focus_status_fp.write("%f \n" % motor_count)
  focus_status_fp.close()


  return()


    
def focus_exit():
  # Clean up and exit

  # Close the motor device
  
  try:
      ch_motor.close()
  except PhidgetException as e:
      print("Phidget Exception %i: %s" % (e.code, e.details))
      print("Press Enter to Exit...\n")
      readin = sys.stdin.read(1)
      exit(1) 

  if diagnostics:
    print("Closed motor device")

  # Close the encoder device

  try:
      ch_encoder.close()
  except PhidgetException as e:
      print("Phidget Exception %i: %s" % (e.code, e.details))
      print("Press Enter to Exit...\n")
      readin = sys.stdin.read(1)
      exit(1) 

  if diagnostics:
    print("Closed encoder device")

  exit(0)
  

# Request a change in the focus setting 
# Return the count change achieved

def change_focus(delta_count):

  # Test for acceptable request and return with no action if out of bounds

  if (abs(delta_count)  > dc_max):
    #Request exceeds 1000 microns (1 mm) limit.  Use a smaller increment.
    return(0)
 
  start_count = motor_count
  
  # Turn on the motor with no motion
  # Acceleration is duty cycles per second per second from 0.5 to 50.
  # Velocity is fraction of duty cycle from -1. to + 1.

  try:
    ch_motor.setAcceleration(2.)
    ch_motor.setTargetVelocity(0.)
    sleep(1) 
  except PhidgetException as e:
    print("Focuser motor Phidget exception %i: %s" % (e.code, e.details))
    print("Exiting ...")
    exit(1)

  delay = abs(delta_count)/counts_per_second

  if (delta_count > 0.):
    try:
      ch_motor.setTargetVelocity(0.9)
      sleep(delay) 
      ch_motor.setTargetVelocity(0.0)
      sleep(1) 
    except PhidgetException as e:
      print("Focuser motor Phidget exception %i: %s" % (e.code, e.details))

  if (delta_count < 0.):
    try:
      ch_motor.setTargetVelocity(-0.9)
      sleep(delay) 
      ch_motor.setTargetVelocity(0.0)
      sleep(1) 
    except PhidgetException as e:
      print("Focuser motor Phidget exception %i: %s" % (e.code, e.details))

  # Write the status file with the focus position in microns

  focus_status_fp = open(focus_status_file,'w')
  focus_status_fp.write("%f \n" % focus_position)
  focus_status_fp.close()
  
  stop_count = motor_count
  actual_delta_count = stop_count - start_count
  
  return(actual_delta_count)  


# Create the root window

root = tk.Tk()
root.title("Focuser")

# Add a frame to the window

mainframe = ttk.Frame(root, padding="3 3 12 12")
mainframe.grid(column=0, row=0, sticky=(tk.N, tk.W, tk.E, tk.S))
mainframe.columnconfigure(0, weight=1)
mainframe.rowconfigure(0, weight=1)

# Create Tk strings for displayed information

tk_focus_position = tk.StringVar()
tk_focus_image = tk.StringVar()
tk_focus_sky = tk.StringVar()

# Initialize the display content strings

tk_focus_position.set(format(focus_position, '7.0f'))
tk_focus_image.set(format(focus_image, '7.1f'))
tk_focus_sky.set(format(focus_sky, '7.1f'))

#Add widgets to the grid within the frame

ttk.Label(mainframe, text="Telescope Focuser").grid(column=1, row=0, sticky=tk.W, columnspan=3)
ttk.Label(mainframe, text="Position").grid(column=0, row=2, sticky=tk.W)
ttk.Label(mainframe, text="microns").grid(column=3, row=2, sticky=tk.W)
ttk.Label(mainframe, text="Image").grid(column=0, row=3, sticky=tk.W)
ttk.Label(mainframe, text="pixels").grid(column=3, row=3, sticky=tk.W)
ttk.Label(mainframe, text="Sky").grid(column=0, row=4, sticky=tk.W)
ttk.Label(mainframe, text="arcseconds").grid(column=3, row=4, sticky=tk.W)

ttk.Label(mainframe, textvariable=tk_focus_position).grid(column=1, row=2, sticky=(tk.W, tk.E))
ttk.Label(mainframe, textvariable=tk_focus_image).grid(column=1, row=3, sticky=(tk.W, tk.E))
ttk.Label(mainframe, textvariable=tk_focus_sky).grid(column=1, row=4, sticky=(tk.W, tk.E))

ttk.Separator(mainframe, orient=tk.HORIZONTAL).grid(row=6, column=0, columnspan=4, sticky=(tk.EW), pady=20)

ttk.Button(mainframe, text="In", command=focus_in).grid(column=0, row=7, sticky=tk.W)
ttk.Button(mainframe, text="Out", command=focus_out).grid(column=1, row=7, sticky=tk.W)
ttk.Button(mainframe, text="Reset", command=focus_reset).grid(column=2, row=7, sticky=tk.W)
ttk.Button(mainframe, text="Exit", command=focus_exit).grid(column=3, row=7, sticky=tk.W)


def callback_change_position(event):

  global change_counts
  global change_position
  global change_image
  global change_sky

  change_position = float(tk_change_position.get())
  change_counts = change_position*counts_per_micron
  change_image = change_counts/counts_per_pixel
  change_sky = change_counts/counts_per_arcsec
  tk_change_image.set(format(change_image,'7.1f'))
  tk_change_sky.set(format(change_sky,'7.1f'))
  return()
  
def callback_change_image(event):

  global change_counts
  global change_position
  global change_image
  global change_sky

  change_image = float(tk_change_image.get())
  change_counts = change_image*counts_per_pixel
  change_sky = change_counts/counts_per_arcsec
  change_position = change_counts/counts_per_micron
  tk_change_position.set(format(change_position,'7.0f'))
  tk_change_sky.set(format(change_sky,'7.1f'))
  return()

def callback_change_sky(event):

  global change_counts
  global change_position
  global change_image
  global change_sky

  change_sky = float(tk_change_sky.get())
  change_counts = change_sky*counts_per_arcsec
  change_image = change_counts/counts_per_pixel
  change_position = change_counts/counts_per_micron
  tk_change_position.set(format(change_position,'7.0f'))
  tk_change_image.set(format(change_image,'7.1f'))
  return()

#   Build initial values for selectors from encoder counts

# How may options?
nlist = 6


# Focuser position in microns
dm = []

# Image plane in pixels
dp = []

# On the sky in arcsecs
da = []

count_increment = counts_per_arcsec*0.5

for i in range(nlist):
  dm.append(format(float(count_increment/counts_per_micron),'7.0f'))
  dp.append(format(float(count_increment/counts_per_pixel),'7.1f'))
  da.append(format(float(count_increment/counts_per_arcsec),'7.1f'))
  count_increment += count_increment

tk_change_position = tk.StringVar()
tk_change_position_selector = ttk.Combobox(mainframe, textvariable=tk_change_position, width=6)
tk_change_position_selector['values'] = tuple(dm)
tk_change_position_selector.grid(column=2, row=2, sticky=tk.W)
tk_change_position_selector.current(2)
tk_change_position_selector.bind("<<ComboboxSelected>>", callback_change_position)

tk_change_image = tk.StringVar()
tk_change_image_selector = ttk.Combobox(mainframe, textvariable=tk_change_image, width=6)
tk_change_image_selector['values'] = tuple(dp)
tk_change_image_selector.grid(column=2, row=3, sticky=tk.W)
tk_change_image_selector.current(2)
tk_change_image_selector.bind("<<ComboboxSelected>>", callback_change_image)


tk_change_sky = tk.StringVar()
tk_change_sky_selector = ttk.Combobox(mainframe, textvariable=tk_change_sky, width=6)
tk_change_sky_selector['values'] = tuple(da)
tk_change_sky_selector.grid(column=2, row=4, sticky=tk.W)
tk_change_sky_selector.current(2)
tk_change_sky_selector.bind("<<ComboboxSelected>>", callback_change_sky)

# Initialize change_counts  based on GUI 

change_position = float(tk_change_position.get())
change_counts = change_position*counts_per_micron

# Pad the widgets for appearance

for child in mainframe.winfo_children(): 
  child.grid_configure(padx=5, pady=5)

#
# GUI mainloop
#


# Tk standard is to use this blocking method 
root.mainloop()


exit(0)
